from rdflib import RDF, RDFS, Graph, URIRef, Literal
from rest_framework.generics import ListAPIView, RetrieveUpdateDestroyAPIView, ListAPIView
from rest_framework import permissions

from django.http import HttpResponse, JsonResponse
from django.views.decorators.csrf import csrf_exempt
from rest_framework.parsers import JSONParser
from reqman.apps.reqtool.rest_api.serializers.requirements import BoilerplateGroupClassesInstancesSerializer, BoilerplateGroupDictionarySerializer

from reqman.apps.reqtool.models.main_req import Main
from reqman.apps.reqtool.rest_api.serializers.main_req import MainSerializer

from reqman.apps.permissions import IsOwnerOrReadOnly
from reqman.apps.reqtool.rest_api.services import fuseki, inference
from reqman.apps.reqtool.rest_api.services.parse_ontologies import Dictionary_Data_list, Ontology_file

from rest_framework.response import Response
from rest_framework.views import APIView

from reqman.apps.reqtool.models.main_req import SHALL_CHOICES, MAIN_SYNTAX

from reqman.apps.reqtool.models.requirements import BoilerplateData, BoilerplateGroupClassesInstances, BoilerplateGroupDictionary
from rest_framework.exceptions import APIException



class MainListAPIView(ListAPIView):
    """
    API view to retrieve list of posts or create new
    """
    permission_classes = [permissions.IsAuthenticatedOrReadOnly]
    serializer_class = MainSerializer
    #
    queryset = Main.objects.all()

    def get_queryset(self):
        group_pk = self.kwargs['boilerplate']
        #print(group_pk)
        queryset_main = Main.objects.filter(boilerplate_of_main = group_pk)
        return queryset_main

    #Custom actions when POST
    def perform_create(self, serializer):
        serializer.save()
        #print the data of the post
        #write_fuseki = fuseki.FusekiActions()
        #write_fuseki.write(serializer.data)

class MainDetailsAPIView(RetrieveUpdateDestroyAPIView):
    """
    API view to retrieve, update or delete post
    """
    permission_classes = [permissions.IsAuthenticatedOrReadOnly]

    serializer_class = MainSerializer
    
    def get_queryset(self):
        group_pk = self.kwargs['boilerplate']
        #print(group_pk)
        queryset_main= Main.objects.filter(boilerplate_of_main = group_pk)
        return queryset_main

    #Custom actions when PUT
    def perform_update(self, serializer):
        instance = serializer.validated_data
        #add data in thn Model BoilerplateData
        es_instance = instance['sys_fun_inter'].rsplit(':', 1)
        ev_instance =instance['verb'].rsplit(' : ', 1)
        eo_instance = instance['flow_function_interface_item_system_state_stateset'].rsplit(':', 1)
        mn = (es_instance[len(es_instance)-1] + ' ' + instance['shall'] + ' ' + ev_instance[len(ev_instance)-1] + ' ' + eo_instance[len(eo_instance)-1])
        boildata = BoilerplateData.objects.filter(owner_data=instance['boilerplate_of_main'].owner, boilerplate_data_id = instance['boilerplate_of_main'])
        boildata.update(main_data = mn )
        #get instances of the ontology that this boilerplate belongs
        data = BoilerplateGroupClassesInstances.objects.filter(classes_instances_owner= instance['boilerplate_of_main'].owner, classes_instances_group_of_boilerplate = boildata.values('group_of_boilerplate_data')[0]['group_of_boilerplate_data'])
        sub_data = data.values('system_choices')[0]['system_choices'] + data.values('function_choices')[0]['function_choices'] + data.values('inteface_choices')[0]['inteface_choices']
        if( instance['sys_fun_inter'] not in sub_data):
            raise APIException("MAIN : Instance of Subject does not exist")
        stateitem_data = data.values('state_choices')[0]['state_choices'] + data.values('item_choices')[0]['item_choices']
        if( instance['state_item_before_verb'] not in stateitem_data):
            raise APIException("MAIN : Instance of State or Item before Verb does not exist")
        statevalue_data = data.values('state_value_choices')[0]['state_value_choices']
        if( instance['statevalue_before_verb'] not in statevalue_data):
            raise APIException("MAIN : Instance of StateValue before Verb does not exist")
        verb_data = data.values('main_verb_choices')[0]['main_verb_choices']
        if( instance['verb'] not in verb_data):
            raise APIException("MAIN : Instance of Verb does not exist")
        quantifier_data = data.values('quantifier_choices')[0]['quantifier_choices']
        if( instance['quantifier'] not in quantifier_data):
            raise APIException("MAIN : Instance of Quantifier does not exist")
        fifuintsysstatset_data =  data.values('system_choices')[0]['system_choices'] + data.values('function_choices')[0]['function_choices'] + data.values('flow_choices')[0]['flow_choices']  + data.values('inteface_choices')[0]['inteface_choices'] + data.values('item_choices')[0]['item_choices'] + data.values('state_choices')[0]['state_choices'] + data.values('state_set_choices')[0]['state_set_choices']
        if( instance['flow_function_interface_item_system_state_stateset'] not in fifuintsysstatset_data):
            raise APIException("MAIN : Instance of Object does not exist")
        stsysconset_data = data.values('system_choices')[0]['system_choices'] + data.values('state_set_choices')[0]['state_set_choices'] + data.values('connection_choices')[0]['connection_choices'] + data.values('state_value_choices')[0]['state_value_choices']
        if( instance['statevalue_system_connection_stateset'] not in stsysconset_data):
            raise APIException("MAIN : Instance of after Object does not exist")
        numunit_data = data.values('number_unit_choices')[0]['number_unit_choices']
        if( instance['mumerical_units'] not in numunit_data):
            raise APIException("MAIN : Instance of Number Unit does not exist")
        #remove requirement if exists in the ontology
        g = Graph()
        g.load(data.values('ontology_file')[0]['ontology_file'], format="turtle")
        rdo_instance= URIRef("http://delab.csd.auth.gr/ontologies/2018/RDO-instances#" + boildata.values('title_data')[0]['title_data'])
        g.remove((rdo_instance, None, None))
        g.remove((None, None, rdo_instance))
        main = Main.objects.filter(boilerplate_of_main = instance['boilerplate_of_main'])
        rdo_instance= URIRef("http://delab.csd.auth.gr/ontologies/2018/RDO-instances#" + boildata.values('title_data')[0]['title_data'] +'_'+ main.values('main_choices')[0]['main_choices'])
        g.remove((rdo_instance, None, None))
        g.remove((None, None, rdo_instance))
        g.serialize(data.values('ontology_file')[0]['ontology_file'], format='turtle')
        #
        curmain_choices = inference.getclassofmain(instance)
        if curmain_choices[0] == 'ERROR':
            raise APIException("Main choices not in [M1-M16]")
        else:
            instance = serializer.save(main_choices = curmain_choices[0])
        #instance = serializer.save()



    #Custom actions when DELETE
    def perform_destroy(self, instance):
        #instance = serializer.validated_data
        

        instance.delete()

class MainChoicesViewSet(ListAPIView):
    """
    API view to retrieve, update or delete
    """
    permission_classes = [permissions.IsAuthenticatedOrReadOnly]
    serializer_class = BoilerplateGroupClassesInstancesSerializer

    def get_queryset(self):
        group_pk = self.kwargs['groupboil']
        queryset_cl= BoilerplateGroupClassesInstances.objects.filter(classes_instances_group_of_boilerplate = group_pk)
        return queryset_cl

class MainChoicesAPIView(APIView):
    def get(self, request, groupboil, instance, comment, format=None):
        data = BoilerplateGroupClassesInstances.objects.filter(classes_instances_group_of_boilerplate = groupboil)
        Dictdata = BoilerplateGroupDictionary.objects.filter(classes_instances_group_of_boilerplate_dic = groupboil)
        
        g = Graph()
        g.load(data.values('ontology_file')[0]['ontology_file'], format="turtle")
        #onotlogy_prefix=self.getontologyfile(g)
        es_instance = instance.rsplit(':')
        #print(Dictdata.values('Dictionary_Data')[0]['Dictionary_Data'])

        class_prefix = self.getClassontologyfile(g,es_instance)
        for s, p, o in g.triples((URIRef( "http://delab.csd.auth.gr/ontologies/2018/RDO-instances#" + es_instance[len(es_instance)-1].strip() ) , RDF.type, URIRef(class_prefix))):
            raise APIException("This instance already exist")
        try:
            g.add(( URIRef( "http://delab.csd.auth.gr/ontologies/2018/RDO-instances#" + es_instance[len(es_instance)-1].strip() ) , RDF.type, URIRef(class_prefix) ))
            g.add(( URIRef( "http://delab.csd.auth.gr/ontologies/2018/RDO-instances#" + es_instance[len(es_instance)-1].strip() ) , RDFS.comment , Literal(comment) ))
            g.serialize(data.values('ontology_file')[0]['ontology_file'], format='turtle')
        except:
            raise APIException("Cannot Add instance")
        tempDict = []
        flag = 0
        for sao in Dictdata.values('Dictionary_Data')[0]['Dictionary_Data']:
            if sao['Name'] == es_instance[0].strip():
                for sao2 in sao['TheClasses']:
                    if sao2['Class'] == es_instance[1].strip():
                        flag = 1
                        for sao3 in sao2['Subclass']:
                            if sao3['SubClass_name'] == es_instance[2].strip():
                                flag = 2
                                if(sao3 ['InstanceOfSubclass'] == ''):
                                    sao3 ['InstanceOfSubclass'] = []
                                    sao3['InstanceOfSubclass'].append({"Instance": es_instance[3].strip(), "Comment": comment })
                                else:
                                    sao3['InstanceOfSubclass'].append({"Instance": es_instance[3].strip(), "Comment": comment })
                        if (flag == 1):
                            if(sao2 ['Instance_of_Class'] == ''):
                                print('sao2')
                                sao2 ['Instance_of_Class'] = []
                                sao2 ['Instance_of_Class'].append({"Instance": es_instance[2].strip(), "Comment": comment })
                            else:
                                sao2 ['Instance_of_Class'].append({"Instance": es_instance[2].strip(), "Comment": comment })
                if(flag==0):
                    sao['GeneralInstances'].append({"Instance": es_instance[1].strip(), "Comment": comment })
            tempDict.append(sao)
        Dictdata.update(Dictionary_Data= tempDict)
        if (es_instance[0].strip() == 'System'):
            m = data.values('system_choices')[0]['system_choices'][:-1]
            if (len(m) ==1 ):
                m+= '["'+instance+'","'+instance+'"]'+']'
            else :
                m+= ',["'+instance+'","'+instance+'"]'+']'
            data.update(system_choices = str(m))
            return HttpResponse('Update Instance')
        if (es_instance[0].strip() == 'Function'):
            m = data.values('function_choices')[0]['function_choices'][:-1]
            if (len(m) ==1 ):
                m+= '["'+instance+'","'+instance+'"]'+']'
            else :
                m+= ',["'+instance+'","'+instance+'"]'+']'
            data.update(function_choices = str(m))
            return HttpResponse('Update Instance')
        if (es_instance[0].strip() == 'Item'):
            m = data.values('item_choices')[0]['item_choices'][:-1]
            if (len(m) ==1 ):
                m+= '["'+instance+'","'+instance+'"]'+']'
            else :
                m+= ',["'+instance+'","'+instance+'"]'+']'
            data.update(item_choices = str(m))
            return HttpResponse('Update Instance')
        if (es_instance[0].strip() == 'Interface'):
            m = data.values('inteface_choices')[0]['inteface_choices'][:-1]
            if (len(m) ==1 ):
                m+= '["'+instance+'","'+instance+'"]'+']'
            else :
                m+= ',["'+instance+'","'+instance+'"]'+']'
            data.update(inteface_choices = str(m))
            return HttpResponse('Update Instance')
        if (es_instance[0].strip() == 'Flow'):
            m = data.values('flow_choices')[0]['flow_choices'][:-1]
            print('data',m)
            if (len(m) ==1 ):
                m+= '["'+instance+'","'+instance+'"]'+']'
            else :
                m+= ',["'+instance+'","'+instance+'"]'+']'
            data.update(flow_choices = str(m))
            print(m)
            return HttpResponse('Update Instance')
        if (es_instance[0].strip() == 'State'):
            m = data.values('state_choices')[0]['state_choices'][:-1]
            if (len(m) ==1 ):
                m+= '["'+instance+'","'+instance+'"]'+']'
            else :
                m+= ',["'+instance+'","'+instance+'"]'+']'
            data.update(state_choices = str(m))
            return HttpResponse('Update Instance')
        if (es_instance[0].strip() == 'StateSet'):
            m = data.values('state_set_choices')[0]['state_set_choices'][:-1]
            if (len(m) ==1 ):
                m+= '["'+instance+'","'+instance+'"]'+']'
            else :
                m+= ',["'+instance+'","'+instance+'"]'+']'
            data.update(state_set_choices = str(m))
            return HttpResponse('Update Instance')
        if (es_instance[0].strip() == 'StateValue'):
            m = data.values('state_value_choices')[0]['state_value_choices'][:-1]
            if (len(m) ==1 ):
                m+= '["'+instance+'","'+instance+'"]'+']'
            else :
                m+= ',["'+instance+'","'+instance+'"]'+']'
            data.update(state_value_choices = str(m))
            return HttpResponse('Update Instance')


        


    def getontologyfile(self, g):
        for ns_prefix, namespace in g.namespaces():
            if ('DSO' in namespace and 'instance' in namespace):
                return namespace
        raise APIException("There is no file for instances")
    
    def getClassontologyfile(self, g, es_instance):
        for ns_prefix, namespace in g.namespaces():
            for s, p, o in g.triples((URIRef( namespace+ es_instance[len(es_instance)-2].strip() ) ,None, None)):
                return URIRef( namespace+ es_instance[len(es_instance)-2].strip() )
        raise APIException("There is no Class in the Ontology ")




        
class ShallChoicesViewSet(APIView):

    def get(self, request):
        response = Response(SHALL_CHOICES)
        return response


class DictionaryViewSet(ListAPIView):
    """
    API view to retrieve list of Sections, Classes, Subclasses, Instances
    """
    permission_classes = [permissions.IsAuthenticatedOrReadOnly]
    serializer_class = BoilerplateGroupDictionarySerializer

    def get_queryset(self):
        group_pk = self.kwargs['groupboil']
        queryset_cl= BoilerplateGroupDictionary.objects.filter(classes_instances_group_of_boilerplate_dic = group_pk)
        return queryset_cl

class MainSyntaxViewSet(APIView):

    def get(self, request):
        response = Response(MAIN_SYNTAX)
        return response


